from typing import Dict

from logging import getLogger
from pathlib import Path

from .storage import load_local
from .storage.plain import ActionCls
from .utils import FileOperation, load_toml

logger = getLogger(__name__)


class SysStorage(ActionCls):
    """
    This is the multi storage, that include multiple storage on the system.
    """

    SYS_CONFIG_PATH: Path = Path("/etc/mynux/config.toml")
    USER_CONFIG_PATH: Path = Path("~/.config/mynux/config.toml").expanduser()

    DEFAULT_SYS_DOTFILES: Path = Path("/etc/mynux/dotfiles")
    DEFAULT_USER_DOTFILES: Path = Path("~/.config/mynux/dotfiles").expanduser()

    DEFAULT_USER_GIT_DIR: Path = Path("~/.config/mynux/repos/git").expanduser()

    DEFAULT_ACTIONS = {"info": False, "file": ["target_dir", "default_file_operation"]}

    def __init__(self, path: Path | None = None, load_sys_conf: bool = True, load_user_conf: bool = True):
        super().__init__()
        self.config: Dict = {}
        if load_sys_conf:
            self._update_conf(self.SYS_CONFIG_PATH)
        if load_user_conf:
            self._update_conf(self.USER_CONFIG_PATH)
        if path:
            self._update_conf(path.resolve())

    def _update_conf(self, path: Path) -> None:
        data = load_toml(path)
        self.config.update(data)

    def action_info(self) -> bool:
        for storage in self.iter_storage():
            print("storage", storage)
            storage.action_info()
        return True

    def action_file(self, target_dir=Path.home(), default_file_operation: FileOperation = FileOperation.LINK) -> bool:
        for storage in self.iter_storage():
            storage.action_file(target_dir, default_file_operation)
        return True

    def iter_storage(self):
        if storage := load_local(self.config.get("system", {}).get("dotfiles") or self.DEFAULT_SYS_DOTFILES):
            yield storage
        if storage := load_local(self.config.get("user", {}).get("dotfiles") or self.DEFAULT_USER_DOTFILES):
            yield storage

        repos_git: Path = self.config.get("user", {}).get("git_dir") or self.DEFAULT_USER_GIT_DIR
        if repos_git.is_dir():
            for path in repos_git.iterdir():
                if storage := load_local(path):
                    yield storage

    def iter_files(self, target_dir: Path | None = None):
        """
        iterate source files (and target files)
        """
        for storage in self.iter_storage():
            yield from storage.iter_files(target_dir)



from typing import Optional

from logging import getLogger
from pathlib import Path

from mynux.utils import FileOperation, check_filter, file_operation_main, iter_gitignore_filter
from mynux.utils.action import ActionCls

logger = getLogger(__name__)


class Storage(ActionCls):
    """
    A simple dotfile directory, no extra mynux.toml file :(
    But good enough to list, copy or link the files.
    """

    DEFAULT_FILTER = [".git/*", "mynux.toml", "*README*"]

    DEFAULT_ACTIONS = {"info": False, "file": ["target_dir", "file_operation"]}

    def __init__(self, path: Path):
        super().__init__()
        self.path = path.resolve()
        if self.path.is_file() and self.path.parent.is_dir():
            self.path = self.path.parent
        self.filters = list(self.iter_filter())

    def __bool__(self):
        return self.path.is_dir()

    def action_file(self, target_dir=Path.home(), default_file_operation: FileOperation = FileOperation.LINK) -> bool:
        """
        copy, link or append files and even more in the mynux class
        """
        for source_file, target_file in self.iter_files(target_dir):
            file_operation_main(source_file, target_file, default_file_operation)
        return True

    def action_info(self) -> bool:
        for file in self.iter_files():
            print(file)
        return True

    def iter_files(self, target_dir: Optional[Path] = None):
        """
        iterate source files (and target files)
        """
        if target_dir is None:
            for file in self.path.glob("**/*"):
                if file.is_file() and not check_filter(file, self.filters):
                    yield file.resolve()
        else:
            target_dir = target_dir.resolve()
            for file in self.path.glob("**/*"):
                if file.is_file() and not check_filter(file, self.filters):
                    yield file.resolve(), target_dir / file.relative_to(self.path)

    def iter_filter(self):
        filters = []
        for name in self.DEFAULT_FILTER:
            filters.append(name)
            yield name

        yield ".gitignore"
        for name in iter_gitignore_filter(self.path / ".gitignore"):
            if name in filters:
                continue
            filters.append(name)
            yield name



import sys
from logging import getLogger
from pathlib import Path

from mynux.storage import load
from mynux.system import SysStorage
from mynux.utils import get_mynux_arg_parser

logger = getLogger(__name__)


def parse_args(*argv: str):
    parser = get_mynux_arg_parser(prog="install")
    parser.add_argument("-t", "--target", type=Path, default=Path.home(), help='target dir, default="~"')
    parser.add_argument("-a", "--action", type=str, choices=["file", "pkgs", "info"], help="just run a single action")
    parser.add_argument("source", nargs="?", type=str, help="source dir")
    return parser, parser.parse_args(argv)


def main(*argv: str) -> int:
    logger.info("run install with: %s", argv)
    _, args = parse_args(*argv)

    storage = load(args.source) or SysStorage()
    if storage is None:
        logger.error('Fail to load storage "%s".', args.source)
        return 1

    match args.action:
        case "file":
            kwargs = {"target_dir": args.target}
        case None:
            kwargs = {"target_dir": args.target}
        case _:
            kwargs = {}

    if storage.action(args.action, **kwargs):
        return 0
    return 1


if __name__ == "__main__":
    sys.exit(main(*sys.argv))



from logging import getLogger
from pathlib import Path

from .mynux import MynuxStorage
from .plain import Storage

logger = getLogger(__name__)


def load(url: str | None) -> Storage | None:
    """
    load storage subclass from single string or return the system storage
    path=kind|content:
        "local|/path/to/dot/dir/or/mynux.toml"
        "git|git@github.com:username/repo.git"
        "git|https://github.com/username/project.git"
        "github|username/project"
    """
    if url is None:
        return None
    match url.split("|", 1):
        case ["local", content]:
            return load_local(Path(content))
        case ["git", content]:
            return load_git(content)
        case _:
            storage = load_local(Path(url))
            if isinstance(storage, Storage):
                return storage
    return None


def load_git(url: str):
    raise NotImplementedError


def load_local(path: Path):
    """
    try to load local storage from path
    (MynuxStorage, PlainStorage)
    """
    for cls in [MynuxStorage, Storage]:
        storage = cls(path)
        if storage:
            return storage
    return None



from typing import Dict, Iterable, List

import subprocess  # nosec
from enum import Enum
from logging import getLogger
from pathlib import Path

from mynux.utils import FileOperation, ask_to_confirm, check_pkg, file_operation_main, load_toml

from .plain import Storage

logger = getLogger(__name__)


class IterPkgAction(Enum):
    ALL = 1
    INSTALLED = 2
    NEW = 3


class MynuxStorage(Storage):
    DEFAULT_ACTIONS = {
        "info": False,
        "file": ["target_dir", "file_operation"],
        "pkgs": True,
    }
    DEFAULT_MYNUX_PATH_NAME: str = "mynux.toml"

    def __init__(self, path: Path):
        super().__init__(path)
        self.mynux_path: Path = path / self.DEFAULT_MYNUX_PATH_NAME
        self.mynux: Dict = load_toml(self.mynux_path)

    def __bool__(self):
        return self.mynux_path.is_file()

    def action_file(self, target_dir=Path.home(), default_file_operation: FileOperation = FileOperation.LINK) -> bool:
        for source_file, target_file in self.iter_files(target_dir):
            relative_source_file = source_file.relative_to(self.path)
            file_operation_str = self.mynux.get("files", {}).get(str(relative_source_file), "none")
            file_operation = FileOperation.get_from_str(file_operation_str, default_file_operation)
            if not file_operation_main(source_file, target_file, file_operation):
                return False
        return True

    def action_pkgs(self) -> bool:
        """
        install the packages from the mynux.toml file
        """
        logger.info("Load new package list.")
        pkgs = list(self.iter_pkgs(IterPkgAction.NEW))
        if pkgs and ask_to_confirm(f"Missing {len(pkgs)} packages. Install them? ", default=True):
            proc = subprocess.run(["sudo", "pacman", "-S", "--noconfirm"] + pkgs)  # nosec
            if proc.returncode != 0:
                logger.error("Fail to install packages! Check your mynux.toml file.")
                return False
            for pkg in pkgs:
                logger.info('Run post install cmds for "%s"', pkg)
                for cmd in self.iter_post_install_cmds(pkg):
                    try:
                        subprocess.run(cmd.split())  # nosec
                    except Exception as exc:
                        logger.exception('Fail cmd "%s"', cmd, exc_info=exc)
                        return False
        return True

    def action_info(self) -> bool:
        total = len(list(self.iter_pkgs()))
        new = len(list(self.iter_pkgs(IterPkgAction.NEW)))
        installed = total - new
        print(f"Packages: {total} | new: {new} | installed: {installed}")
        return True

    def iter_post_install_cmds(self, pkg: str):
        cmds = self.mynux.get("post_install_cmd", {}).get(pkg, [])
        if isinstance(cmds, list):
            yield from cmds
        if isinstance(cmds, str):
            yield cmds

    def iter_pkgs(self, pkg_action: IterPkgAction = IterPkgAction.ALL) -> Iterable:
        pkgs: List[str] = list(self.mynux.get("package", []))
        match pkg_action:
            case IterPkgAction.ALL:
                return iter(pkgs)
            case IterPkgAction.INSTALLED:
                return filter(check_pkg, pkgs)
            case IterPkgAction.NEW:
                return filter(lambda p: not check_pkg(p), pkgs)
            case _:
                raise Exception("WTF, how is this happening? Do you append the IterPkgAction?")



import sys
from logging import getLogger

from mynux.system import SysStorage
from mynux.utils.__init__ import get_mynux_arg_parser

from ..storage import load

logger = getLogger(__name__)


def main(*argv: str) -> int:
    logger.info("run info with: %s", argv)
    parser = get_mynux_arg_parser(prog="info")
    parser.add_argument("source", nargs="?", type=str, help="source dir")

    args = parser.parse_args(argv)
    storage = load(args.source) or SysStorage()
    return int(storage.action("info"))


if __name__ == "__main__":
    sys.exit(main(*sys.argv))



from typing import Dict, List

from logging import getLogger

logger = getLogger(__name__)


class ActionCls:
    """
    just handel the action calls
    """

    DEFAULT_ACTIONS: Dict[str, List | bool] = {}

    def __call__(self, name: str | None = None, **kwargs) -> bool:
        try:
            return self.action(name, **kwargs)
        except Exception as exc:
            logger.exception('Fail to call "%s".', self, exc_info=exc)
        return False

    def check_action(self, name: str) -> bool:
        return hasattr(self, f"action_{name}")

    def action(self, name: str | None = None, **kwargs) -> bool:
        """execute a single action or all DEFAULT_ACTIONS"""
        if name is None:
            for action, args in self.DEFAULT_ACTIONS.items():
                action_kwargs = {}
                match args:
                    case list():
                        action_kwargs = {arg: kwargs.get(arg) for arg in args if arg in kwargs}
                    case False:
                        continue
                # exit if any action fail
                if not self.action(action, **action_kwargs):
                    return False
            return True

        func = getattr(self, f"action_{name}", None)
        if func:
            return func(**kwargs)
        logger.warning('The action "%s" is not available in the class "%s"!', name, self.__class__)
        return False



from typing import Optional

import subprocess  # nosec
from logging import getLogger
from pathlib import Path

from .plain import Storage

logger = getLogger(__name__)


class RemoteGitStorage(Storage):
    def __init__(self, path: Path, url: Optional[str]):
        super().__init__(path)
        self.url = url or ""  # TODO: load url from .git

    def clone(self, path: Path):
        subprocess.run(["git", "clone", self.url, path])  # nosec
        return RemoteGitStorage(path, self.url)



import sys
from argparse import REMAINDER, SUPPRESS, ArgumentParser
from logging import basicConfig, getLogger

from mynux import __version__, cmds

logger = getLogger(__name__)


def help_cmds():
    width = max(map(len, cmds.keys()))
    for name, cmd in cmds.items():
        func = cmd.load()
        help_str = func.__doc__ or ""
        print(f"{name:>{width}} --> {help_str}")


def main_parse_args(argv=None):
    parser = ArgumentParser()
    parser.add_argument("-V", "--version", action="version", version=__version__)
    parser.add_argument("-v", "--verbose", action="count", help="verbose level... repeat up to three times")
    parser.add_argument("-H", "--help_cmds", action="store_true", help="some command info")
    parser.add_argument("cmd", nargs="?", choices=cmds.keys(), help="select one command")
    parser.add_argument("args", help=SUPPRESS, nargs=REMAINDER)
    return parser, parser.parse_args(argv)


def main(argv=None):
    parser, args = main_parse_args(argv)

    if args.verbose:
        level = 40 - args.verbose * 10 if args.verbose <= 3 else 30
        log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        basicConfig(level=level, format=log_format)

    if args.help_cmds:
        return help_cmds()

    if args.cmd:
        try:
            func = cmds[args.cmd].load()
            sys.exit(func(*args.args))
        except Exception as exc:
            if args.verbose:
                raise
            logger.error('Oh no, a error :(\nError: "%s"', exc)
            logger.error("Run with --verbose for more information.")
            sys.exit(1)

    sys.exit(parser.print_help())


if __name__ == "__main__":
    main()



from typing import Dict

import sys

from mynux.utils.__init__ import iter_cmds

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


__version__: str = get_version()

cmds = dict(iter_cmds())

config: Dict[str, Dict] = {}



from typing import Dict

import fnmatch
import re
import shutil
import subprocess  # nosec
import sys
from argparse import ArgumentParser
from enum import Enum
from importlib.metadata import entry_points
from logging import getLogger
from pathlib import Path

if sys.version_info >= (3, 11):
    import tomllib  # pylint: disable=import-error
else:
    import tomli as tomllib

logger = getLogger(__name__)


class FileOperation(Enum):
    LINK = 1
    COPY = 2
    APPEND = 3

    @classmethod
    def get_from_str(cls, name: str, default=None):
        match name.lower():
            case "link" | "ln":
                return cls.LINK
            case "copy" | "cp":
                return cls.COPY
            case "append" | "add" | "+":
                return cls.APPEND
        return default


class UserAction(Enum):
    ASK = 1
    DELETE = 2
    SKIPP = 3


def get_mynux_arg_parser(prog: str) -> ArgumentParser:
    return ArgumentParser(prog=f"mynux {prog}")


def ask_to_confirm(msg: str, default: bool = True) -> bool:
    """ask user boolean question"""
    if default:
        result = input(msg + "[Y/n] ") or "y"
    else:
        result = input(msg + "[y/N] ") or "n"
    return result.lower() in ["y", "yes"]


def iter_cmds():
    for entry_point in entry_points().select(group="mynux.cmd"):
        yield entry_point.name, entry_point


def check_filter(path, filters):
    """return True if som filter match"""
    for pattern in filters:
        if re.search(fnmatch.translate(pattern), str(path)):
            return True
    return False


def iter_gitignore_filter(gitignore_path: Path):
    """load filters from gitignore"""
    if gitignore_path.is_file():
        with gitignore_path.open() as file:
            gitignore_filters = file.read().splitlines()
        for name in gitignore_filters:
            if not name or name.startswith("#"):
                continue
            yield name


def check_target_file(target_file: Path, user_action: UserAction = UserAction.ASK) -> bool:
    """
    Check if the target file exist. Return only True if file did not exist.
    If the target file exist it could be deleted:
    UserAction.ASK -> the user can decide
              .DELETE -> deleting file
              .SKIPP -> return False, file would be there
    """
    if not target_file.is_file() and not target_file.is_symlink():
        return True

    match user_action:
        case UserAction.ASK:
            if (input(f'File "{target_file}" exist! Overwrite? [Y/n] ') or "Y") in ["y", "Y", "yes"]:
                target_file.unlink()
            else:
                logger.info('skipp file "%s" by user', target_file)
                return False
        case UserAction.DELETE:
            target_file.unlink()
        case UserAction.SKIPP:
            return False
        case _:
            raise Exception("WTF, how is this happening? Do you append the UserAction?")

    # just to make sure that the file is deleted
    return not target_file.is_file() and not target_file.is_symlink()


def file_operation_link(source_file: Path, target_file: Path, user_action: UserAction = UserAction.ASK) -> bool:
    if not check_target_file(target_file, user_action):
        return False
    try:
        target_file.parent.mkdir(parents=True, exist_ok=True)
        target_file.symlink_to(source_file)
        return True
    except Exception as exc:
        logger.error("Error at file_operation_link", exc_info=exc)
        return False


def file_operation_copy(source_file: Path, target_file: Path, user_action: UserAction = UserAction.ASK) -> bool:
    if not check_target_file(target_file, user_action):
        return False
    try:
        target_file.unlink(missing_ok=True)
        target_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(source_file, target_file)
        return True
    except Exception as exc:
        logger.error("Error at file_operation_copy", exc_info=exc)
        return False


def file_operation_append(source_file: Path, target_file: Path) -> bool:
    raise NotImplementedError


def file_operation_main(source_file: Path, target_file: Path, file_operation: FileOperation) -> bool:
    match file_operation:
        case FileOperation.LINK:
            result = file_operation_link(source_file, target_file)
        case FileOperation.COPY:
            result = file_operation_copy(source_file, target_file)
        case FileOperation.APPEND:
            result = file_operation_append(source_file, target_file)
        case _:
            result = False
    if result:
        logger.info('Install file "%s" to "%s".', source_file, target_file)
    else:
        logger.warning('Fail to install file "%s" to "%s".', source_file, target_file)
    return result


def check_pkg(pkg: str) -> bool:
    """return True if package is installed"""
    proc = subprocess.run(["pacman", "-Qn", pkg], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec
    if proc.returncode == 0:
        return True

    proc = subprocess.run(["pacman", "-Qg", pkg], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)  # nosec
    group_pkg = [line.decode().split()[1] for line in proc.stdout.splitlines()]
    return proc.returncode == 0 and all(map(check_pkg, group_pkg))


def load_toml(path: Path) -> Dict:
    if not path.is_file():
        return {}
    try:
        with path.open("rb") as file:
            return tomllib.load(file)
    except Exception as exc:
        logger.exception('Fail to load "%s".', path, exc_info=exc)
    return {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mynux', 'mynux.cli', 'mynux.storage', 'mynux.utils']

package_data = \
{'': ['*']}

install_requires = \
['tox>=3.25.1,<4.0.0']

extras_require = \
{':python_version < "3.11"': ['tomli>=2.0.1,<3.0.0']}

entry_points = \
{'console_scripts': ['mynux = mynux.__main__:main'],
 'mynux.cmd': ['info = mynux.cli.info:main',
               'install = mynux.cli.install:main']}

setup_kwargs = {
    'name': 'mynux',
    'version': '0.2.0',
    'description': '',
    'long_description': 'mynux - My Linux\n================\n!!!Work in progress!!!\n\nJust a dotfile manager with some extras.\n\nInstall\n-------\nIt\'s a python package, so simple install with pip::\n\n    pip install mynux\n\nor get the latest version from git::\n\n    pip install git+https://github.com/axju/mynux.git\n\nQuickstart\n----------\nIf you only want to install a single dotfile directory to your home directory,\nyou can run::\n\n    mynux install path/to/directory/or/git/url\n\nAny repo will be clone into "~/.config/mynux/repos/". After the files will\nbe linked or copy to your home directory. With the mynux.toml file you can also\ninstall package.\n\nMulti storage setup\n~~~~~~~~~~~~~~~~~~~\nAdd a repo or directory to the local config directory::\n\n    mynux add git@github.com/axju/dotfiles.git\n    mynux add /path/to/dir\n    mynux add --name my-dotfile git@github.com/axju/dotfiles.git\n\nThe command::\n\n    mynux install\n\nwill install all mynux storage from the mynux config directory. To change the order\nyou can uses the argument "--sort=name1,name2..." or change the config file\n"~./config/mynux/config.toml".\n\nDev\n---\nThis project is mange with `Poetry <https://python-poetry.org/>`_. First install\nPoetry, then run::\n\n    poetry install\n\nto install the project. Setup the per-commit hook with::\n\n    poetry run pre-commit install\n\nNow you are ready to develop. We have also a "Makefile". Look at the file to see\nthe bare commands. This ar some of them::\n\n    make install\n    make formatting\n    make test-all\n\nBugs\n----\nIf you find any bugs, pleas open an `Issue <https://github.com/axju/mynux/issues/new>`_.\n',
    'author': 'axju',
    'author_email': 'moin@axju.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
